# :coding: utf-8
# :copyright: Copyright (c) 2013 Martin Pengelly-Phillips
# :license: See LICENSE.txt.

import re

import pytest

import clique


def test_assemble():
    '''Assemble collections from arbitrary items.'''
    items = [
        'file.ext',
        'single.1.ext',
        '1', '3',
        '001', '003',
        'head.001.tail', 'head.002.tail',
        'head.1.tail', 'head.2.tail',
        'head.001.tail3', 'head.002.tail3',
        'head_v01.tail', 'head_v02.tail',
        'head_v1.tail', 'head_v2.tail',
        'head1_010_v1.0001.tail', 'head1_010_v1.0002.tail',
        'head1_010_v2.0001.tail', 'head1_010_v2.0002.tail',
        'head1_020_v1.0001.tail', 'head1_020_v1.0002.tail'
    ]
    
    collections, remainder = clique.assemble(items)

    expected = [
        clique.Collection('', '', 0, indexes=set([1, 3])),
        clique.Collection('', '', 3, indexes=set([1, 3])),
        clique.Collection('head.', '.tail', 3, indexes=set([1, 2])),
        clique.Collection('head.', '.tail', 0, indexes=set([1, 2])),
        clique.Collection('head.', '.tail3', 3, indexes=set([1, 2])),

        clique.Collection('head_v', '.tail', 2, indexes=set([1, 2])),
        clique.Collection('head_v', '.tail', 0, indexes=set([1, 2])),

        clique.Collection('head1_010_v1.', '.tail', 4, indexes=set([1, 2])),
        clique.Collection('head1_010_v2.', '.tail', 4, indexes=set([1, 2])),
        clique.Collection('head1_020_v1.', '.tail', 4, indexes=set([1, 2])),
        clique.Collection('head1_010_v', '.0001.tail', 0, indexes=set([1, 2])),
        clique.Collection('head1_010_v', '.0002.tail', 0, indexes=set([1, 2])),
        clique.Collection('head1_', '_v1.0001.tail', 3, indexes=set([10, 20])),
        clique.Collection('head1_', '_v1.0002.tail', 3, indexes=set([10, 20]))
    ]

    assert sorted(collections) == sorted(expected)

    expected = ['file.ext', 'single.1.ext']
    assert sorted(remainder) == sorted(expected)


def test_assemble_case_sensitive():
    '''Assemble collections respecting casing.'''
    collections, _ = clique.assemble(
        [
            'head_v1.001.ext', 'head_v1.002.ext',
            'HEAD_v1.003.ext', 'HEAD_v1.004.ext'
        ],
        case_sensitive=True
    )
    expected = [
        clique.Collection('head_v1.', '.ext', 3, indexes=set([1, 2])),
        clique.Collection('HEAD_v1.', '.ext', 3, indexes=set([3, 4]))
    ]

    assert sorted(collections) == sorted(expected)


def test_assemble_case_insensitive():
    '''Assemble collections ignoring casing.'''
    collections, _ = clique.assemble(
        ['head_v1.001.ext', 'HEAD_v1.002.ext', 'head_v1.003.ext'],
        case_sensitive=False
    )
    expected = [
        clique.Collection('head_v1.', '.ext', 3, indexes=set([1, 2, 3]))
    ]

    assert collections == expected


def test_assemble_no_patterns():
    '''Assemble with no patterns.'''
    assert clique.assemble(['1', '2'], patterns=[]) == ([], ['1', '2'])


def test_assemble_with_custom_pattern():
    '''Assemble with custom pattern.'''
    collections, _ = clique.assemble(
        ['head_v1.001.ext', 'head_v1.002.ext'],
        patterns=[re.compile(r'\.{0}\.ext$'.format(clique.DIGITS_PATTERN))]
    )
    expected = [clique.Collection('head_v1.', '.ext', 3, indexes=set([1, 2]))]

    assert collections == expected


def test_assemble_compile_pattern():
    '''Compile string based custom pattern.'''
    collections, _ = clique.assemble(
        ['head_v1.001.ext', 'head_v1.002.ext'],
        patterns=[r'\.{0}\.ext$'.format(clique.DIGITS_PATTERN)]
    )
    expected = [clique.Collection('head_v1.', '.ext', 3, indexes=set([1, 2]))]

    assert collections == expected


def test_assemble_minimum_items_filter():
    '''Filter assembled collections by minimum number of items.'''
    items = [
        'head_v1.001.ext', 'head_v1.002.ext', 'head_v1.003.ext',
        'head_v2.001.ext', 'head_v2.002.ext'
    ]

    collections, _ = clique.assemble(items, minimum_items=1)
    expected = [
        clique.Collection('head_v', '.001.ext', 0, indexes=set([1, 2])),
        clique.Collection('head_v', '.002.ext', 0, indexes=set([1, 2])),
        clique.Collection('head_v', '.003.ext', 0, indexes=set([1])),
        clique.Collection('head_v1.', '.ext', 3, indexes=set([1, 2, 3])),
        clique.Collection('head_v2.', '.ext', 3, indexes=set([1, 2]))
    ]
    assert sorted(collections) == sorted(expected)

    collections, _ = clique.assemble(items, minimum_items=3)
    expected = [
        clique.Collection('head_v1.', '.ext', 3, indexes=set([1, 2, 3]))
    ]
    assert collections == expected

    collections, _ = clique.assemble(items, minimum_items=5)
    expected = []
    assert collections == expected


@pytest.mark.parametrize(('items', 'expected'), [
    (
        ['0999', '1000', '1001', '9999'],
        [
            clique.Collection('', '', 4, indexes=set([999, 1000, 1001, 9999]))
        ]
    ),
    (
        ['999', '0999', '1000', '1001', '9999'],
        [
            clique.Collection('', '', 4, indexes=set([999, 1000, 1001, 9999])),
            clique.Collection('', '', 0, indexes=set([999, 1000, 1001, 9999]))
        ]
    ),
    (
        ['0999', '1000', '9999', '10001'],
        [
            clique.Collection('', '', 4, indexes=set([999, 1000, 9999])),
            clique.Collection('', '', 0, indexes=set([1000, 9999, 10001]))
        ]
    )
], ids=[
    'padding boundary aligned',
    'ambiguous padding boundary',
    'non-padding boundary indexes'
])
def test_assemble_boundary_padding(items, expected):
    '''Assemble items across a padding boundary.'''
    collections, _ = clique.assemble(items)
    assert sorted(collections) == sorted(expected)


def test_assemble_remainder_has_no_duplicates():
    '''Assemble items and check remainder contains no duplicates.'''
    items = ['00-11-22-33-44-55.jpg']
    collections, remainder = clique.assemble(items, minimum_items=2)
    assert remainder == items


@pytest.mark.parametrize(('items', 'assume_padded', 'expected'), [
    (
        ['1000', '1001', '1002'],
        False,
        [
            clique.Collection('', '', 0, indexes={1000, 1001, 1002})
        ]
    ),
    (
        ['1000', '1001', '1002'],
        True,
        [
            clique.Collection('', '', 4, indexes={1000, 1001, 1002})
        ]
    ),
    (
        ['999', '1000', '1001'],
        False,
        [
            clique.Collection('', '', 0, indexes={999, 1000, 1001})
        ]
    ),
    (
        ['999', '1000', '1001'],
        True,
        [
            clique.Collection('', '', 0, indexes={999, 1000, 1001})
        ]
    ),
    (
        ['0999', '1000', '1001'],
        False,
        [
            clique.Collection('', '', 4, indexes={999, 1000, 1001})
        ]
    ),
    (
        ['0999', '1000', '1001'],
        True,
        [
            clique.Collection('', '', 4, indexes={999, 1000, 1001})
        ]
    ),
    (
        [],
        True,
        []
    )
], ids=[
    'ambiguous assume unpadded',
    'ambiguous assume padded',
    'unambiguous unpadded assume unpadded',
    'unambiguous unpadded assume padded',
    'unambiguous padded assume unpadded',
    'unambiguous padded assume padded',
    'empty collection'
])
def test_assemble_assume_padded(items, assume_padded, expected):
    '''Assemble items according to assume padded option.'''
    collections, remainder = clique.assemble(
        items, assume_padded_when_ambiguous=assume_padded
    )
    assert collections == expected


@pytest.mark.parametrize(('value', 'pattern', 'expected'), [
    ('/path/to/file.%04d.ext []', None,
     clique.Collection('/path/to/file.', '.ext', 4, [])),
    ('/path/to/file.%04d.ext [1-3, 5, 7-8]', None,
     clique.Collection('/path/to/file.', '.ext', 4, [1, 2, 3, 5, 7, 8])),
    ('/path/to/file.%d.ext [1-3, 5, 7-8]', None,
     clique.Collection('/path/to/file.', '.ext', 0, [1, 2, 3, 5, 7, 8])),
    ('/path/to/file.%d.ext 1-8',
     '{head}{padding}{tail} {range}',
     clique.Collection('/path/to/file.', '.ext', 0, [1, 2, 3, 4, 5, 6, 7, 8])),
    ('/path/to/file.%d.ext 1-8 [2, 4-6]',
     '{head}{padding}{tail} {range} [{holes}]',
     clique.Collection('/path/to/file.', '.ext', 0, [1, 3, 7, 8]))
], ids=[
    'empty',
    'padded',
    'unpadded',
    'custom range pattern',
    'custom holes pattern'
])
def test_parse(value, pattern, expected):
    '''Construct collection by parsing formatted string.'''
    if pattern is None:
        assert clique.parse(value) == expected
    else:
        assert clique.parse(value, pattern=pattern) == expected


def test_non_matching_parse():
    '''Fail to parse non-matching value.'''
    with pytest.raises(ValueError):
        clique.parse('')


# :coding: utf-8
# :copyright: Copyright (c) 2013 Martin Pengelly-Phillips
# :license: See LICENSE.txt.

'''Clique documentation build configuration file'''

import os
import re
import sys

# -- General ------------------------------------------------------------------

# Inject source onto path so that autodoc can find it by default, but in such a
# way as to allow overriding location.
sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'source'))
)

# Extensions
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'lowdown'
]

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Clique'
copyright = u'2013, Martin Pengelly-Phillips'

# Version
with open(
    os.path.join(
        os.path.dirname(__file__), '..', 'source', 'clique', '_version.py'
    )
) as _version_file:
    _version = re.match(
        r'.*__version__ = \'(.*?)\'', _version_file.read(), re.DOTALL
    ).group(1)

version = _version
release = _version

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# A list of prefixes to ignore for module listings
modindex_common_prefix = ['clique.']

# -- HTML output --------------------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named 'default.css' will overwrite the builtin 'default.css'.
html_static_path = []

# If True, copy source rst files to output for reference
html_copy_source = True

# -- Autodoc ------------------------------------------------------------------

autodoc_default_options = {
    'members': True,
    'undoc-members': True,
    'show-inheritance': True
}

autodoc_member_order = 'bysource'


def autodoc_skip(app, what, name, obj, skip, options):
    '''Don't skip __init__ method for autodoc.'''
    if name == '__init__':
        return False

    return skip


# -- Intersphinx --------------------------------------------------------------

intersphinx_mapping = {
    'python': ('http://docs.python.org/', None),
    'pip': ('https://pip.pypa.io/en/stable/', None)
}


# -- Setup --------------------------------------------------------------------

def setup(app):
    app.connect('autodoc-skip-member', autodoc_skip)



# :coding: utf-8
# :copyright: Copyright (c) 2013 Martin Pengelly-Phillips
# :license: See LICENSE.txt.

import pytest

from clique.sorted_set import SortedSet


def test_str():
    '''String representation.'''
    sorted_set = SortedSet([1, 3, 2])
    assert str(sorted_set) == '[1, 2, 3]'


def test_repr():
    '''Repr representation.'''
    sorted_set = SortedSet([1, 3, 2])
    assert repr(sorted_set) == '<SortedSet "[1, 2, 3]">'


@pytest.mark.parametrize(('sorted_set', 'item', 'expected'), [
    (SortedSet([1]), 1, True),
    (SortedSet([1]), 10, False)
], ids=[
    'item present',
    'item not present'
])
def test_contains(sorted_set, item, expected):
    '''Check item membership.'''
    assert (item in sorted_set) is expected


@pytest.mark.parametrize(('sorted_set', 'expected'), [
    (SortedSet(), 0),
    (SortedSet([]), 0),
    (SortedSet([1]), 1),
    (SortedSet([1, 2, 3]), 3),
    (SortedSet([1, 1, 2, 2, 3, 3]), 3)
], ids=[
    'no iterable',
    'empty iterable',
    'single item',
    'multiple items',
    'duplicate multiple items'
])
def test_len(sorted_set, expected):
    '''Calculate set length.'''
    assert len(sorted_set) == expected


@pytest.fixture
def standard_set(request):
    '''Return sorted set.'''
    return SortedSet([4, 5, 6, 7, 2, 1, 1])


@pytest.mark.parametrize(('sorted_set', 'item', 'expected'), [
    (SortedSet(), 1, 1),
    (SortedSet([1]), 1, 1),
    (SortedSet([1]), 2, 2)
], ids=[
    'item',
    'existing item',
    'new item'
])
def test_add(sorted_set, item, expected):
    '''Add item.'''
    sorted_set.add(item)
    assert item in sorted_set
    assert len(sorted_set) == expected


@pytest.mark.parametrize(('sorted_set', 'item'), [
    (SortedSet([1]), 1),
    (SortedSet(), 1)
], ids=[
    'present item',
    'missing item'
])
def test_discard(sorted_set, item):
    '''Discard item.'''
    sorted_set.discard(item)
    assert item not in sorted_set


def test_update():
    '''Update with items from another iterable.'''
    sorted_set = SortedSet([1, 2, 3])
    sorted_set.update([1, 4, 5, 6])
    assert list(sorted_set) == [1, 2, 3, 4, 5, 6]


def test_iter():
    '''Iterate over ordered items.'''
    sorted_set = SortedSet([5, 2, 3, 4, 8, 9, 1, 1])
    assert list(sorted_set) == [1, 2, 3, 4, 5, 8, 9]




# :coding: utf-8
# :copyright: Copyright (c) 2013 Martin Pengelly-Phillips
# :license: See LICENSE.txt.

import os
import re

from setuptools import setup, find_packages

ROOT_PATH = os.path.dirname(os.path.realpath(__file__))
SOURCE_PATH = os.path.join(ROOT_PATH, 'source')
README_PATH = os.path.join(ROOT_PATH, 'README.rst')

PACKAGE_NAME = 'clique'

# Read version from source.
with open(
    os.path.join(SOURCE_PATH, PACKAGE_NAME, '_version.py')
) as _version_file:
    VERSION = re.match(
        r'.*__version__ = \'(.*?)\'', _version_file.read(), re.DOTALL
    ).group(1)


# Compute dependencies.
INSTALL_REQUIRES = []
DOC_REQUIRES = [
    'sphinx >= 2, < 4',
    'sphinx_rtd_theme >= 0.1.6, < 1',
    'lowdown >= 0.2.0, < 1'
]
TEST_REQUIRES = [
    'pytest-runner >= 2.7, < 3',
    'pytest >= 2.3.5, < 5',
    'pytest-cov >= 2, < 3',
    'six > 1.14.0, < 2'
]

# Readthedocs requires Sphinx extensions to be specified as part of
# install_requires in order to build properly.
if os.environ.get('READTHEDOCS', None) == 'True':
    INSTALL_REQUIRES.extend(DOC_REQUIRES)


setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description='Manage collections with common numerical component',
    long_description=open(README_PATH).read(),
    keywords='sequence, pattern, filesystem, collection, numerical',
    url='https://gitlab.com/4degrees/clique',
    author='Martin Pengelly-Phillips',
    author_email='martin@4degrees.ltd.uk',
    license='Apache License (2.0)',
    packages=find_packages(SOURCE_PATH),
    package_dir={
        '': 'source'
    },
    install_requires=INSTALL_REQUIRES,
    extras_require={
        'doc': DOC_REQUIRES,
        'test': TEST_REQUIRES,
        'dev': DOC_REQUIRES + TEST_REQUIRES
    },
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3'
    ],
    zip_safe=False,
    python_requires=">=2.7, <4.0"
)
